# Copyright (c) OpenMMLab. All rights reserved.
from .download import download
from .gridsearch import gridsearch
from .install import install
from .list import list_package
from .run import run
from .search import get_model_info
from .test import test
from .train import train
from .uninstall import uninstall

__all__ = [
    'download', 'install', 'list_package', 'uninstall', 'train', 'test', 'run',
    'gridsearch', 'download', 'get_model_info'
]
